# This file may not be shared/redistributed without permission. Please read copyright notice in the git repo. If this file contains other copyright notices disregard this text.
from gymnasium.wrappers import TimeLimit

from irlc.pacman.pacman_environment import PacmanWinWrapper




class AsyncPacmanWinWrapper(PacmanWinWrapper):
    async def async_step(self, action):
        observation, reward, done, truncated, info = await self.env.async_step(action)
        if self.env.unwrapped.game.state.is_won():
            reward = 1
        else:
            reward = 0
        return observation, reward, done, truncated, info


class AsyncTimeLimit(TimeLimit):
    # def __init__(s

    async def async_step(self, action):
        """Steps through the environment and if the number of steps elapsed exceeds ``max_episode_steps`` then truncate.

        Args:
            action: The environment step action

        Returns:
            The environment step ``(observation, reward, terminated, truncated, info)`` with `truncated=True`
            if the number of steps elapsed >= max episode steps

        """
        observation, reward, terminated, truncated, info = await self.env.async_step(action)
        self._elapsed_steps += 1

        if self._elapsed_steps >= self._max_episode_steps:
            truncated = True

        return observation, reward, terminated, truncated, info
